package net.astah.stpa.stamp.notation.edit.command;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.common.command.CompoundCommand;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.edit.domain.EditingDomain;

import com.google.common.collect.Lists;

import net.astah.dd.di.util.DiagramElements;
import net.astah.emf.edit.command.SynchronizeListCommand;
import net.astah.stpa.stamp.NamedElement;
import net.astah.stpa.stamp.ProcessModel;
import net.astah.stpa.stamp.ProcessValue;
import net.astah.stpa.stamp.ProcessVariable;
import net.astah.stpa.stamp.notation.STAMPCompartment;
import net.astah.stpa.stamp.notation.STAMPDiagramElement;
import net.astah.stpa.stamp.notation.STAMPLabel;
import net.astah.stpa.stamp.notation.STAMPNotationFactory;

/**
 * Process Model に Process Variable や Process Value
 * が追加／削除／順序変更された時に図要素の同期処理を行うコマンドです。
 */
public class SynchronizeProcessModelCommand extends CompoundCommand {

    private final EditingDomain domain;

    private final STAMPCompartment owner;

    /**
     * コンストラクタ
     *
     * @param domain
     *            コマンドを実行するドメイン
     * @param owner
     *            同期対象となる図要素の owner 。 Process Model に対応する図要素です。
     */
    public SynchronizeProcessModelCommand(final EditingDomain domain,
            final STAMPCompartment owner) {
        this.domain = domain;
        this.owner = owner;
    }

    @Override
    protected boolean prepare() {
        prepareCommand();
        return super.prepare();
    }

    private void prepareCommand() {
        if (domain == null || owner == null || owner.getModelElements().isEmpty()) {
            return;
        }
        final List<STAMPDiagramElement> elements = getElementsInCompartment(
                owner.getElementInCompartments(),
                DiagramElements.getModelElement(owner, ProcessModel.class).getVariables());
        final List<STAMPDiagramElement> elementsWithCompartmentName = Lists
                .newArrayList(owner.getSTAMPOwnedElements().get(0));
        elementsWithCompartmentName.addAll(elements);
        // この順序でないとProcess Variableの追加で例外が発生する
        append(new SynchronizeListCommand(domain, owner.getOwnedElements(),
                elementsWithCompartmentName));
        append(new SynchronizeListCommand(domain, owner.getElementInCompartments(), elements));
    }

    private List<STAMPDiagramElement> getElementsInCompartment(
            final EList<STAMPDiagramElement> oldElements,
            final EList<ProcessVariable> processVariables) {
        final List<STAMPDiagramElement> newElements = new ArrayList<>();
        for (final ProcessVariable processVariable : processVariables) {
            final STAMPDiagramElement variableLabel = getDiagramElementByNamedElement(oldElements,
                    processVariable);
            newElements.add(variableLabel != null ? variableLabel : createLabel(processVariable));
            for (final ProcessValue processValue : processVariable.getValues()) {
                final STAMPDiagramElement valueLabel = getDiagramElementByNamedElement(oldElements,
                        processValue);
                newElements.add(valueLabel != null ? valueLabel : createLabel(processValue));
            }
        }
        return newElements;
    }

    private STAMPDiagramElement getDiagramElementByNamedElement(
            final EList<STAMPDiagramElement> elements, final NamedElement element) {
        return elements.stream().filter(e -> e.getModelElements().contains(element)).findFirst()
                .orElse(null);
    }

    private STAMPLabel createLabel(final NamedElement element) {
        final STAMPLabel label = STAMPNotationFactory.eINSTANCE.createSTAMPLabel();
        label.getModelElements().add(element);
        label.setText(element.getName());
        return label;
    }
}
