package net.astah.stpa.stamp.notation.edit.reactor;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.command.CompoundCommand;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.edit.command.SetCommand;
import org.eclipse.emf.edit.domain.EditingDomain;

import com.google.common.collect.FluentIterable;
import com.google.common.collect.Iterables;

import net.astah.dd.dc.Dimension;
import net.astah.dd.di.DIPackage;
import net.astah.dd.di.DiagramElement;
import net.astah.dd.di.edit.provider.IItemGeometryProvider;
import net.astah.dd.di.edit.provider.LayoutCalculator;
import net.astah.dd.di.util.DiagramElements;
import net.astah.emf.common.notify.IPathNotification;
import net.astah.emf.edit.command.SynchronizeListCommand;
import net.astah.emf.edit.notify.EFeaturePathReactor;
import net.astah.emf.edit.provider.ItemProviders;
import net.astah.notation.Location;
import net.astah.notation.LocationConstraint;
import net.astah.notation.NotationFactory;
import net.astah.notation.NotationPackage;
import net.astah.stpa.stamp.Link;
import net.astah.stpa.stamp.Signal;
import net.astah.stpa.stamp.notation.STAMPEdge;
import net.astah.stpa.stamp.notation.STAMPLabel;
import net.astah.stpa.stamp.notation.STAMPNotationFactory;

/**
 * CA/FB のモデルが追加された時に図要素の同期処理を行うクラスです。
 *
 */
public class LinkEdgeSynchronizer implements EFeaturePathReactor {

    /**
     * @see EFeaturePathReactor#process(EditingDomain, Collection)
     */
    @Override
    public Command process(EditingDomain domain, Collection<IPathNotification> notifications) {
        final Set<Object> processed = new HashSet<>();
        final CompoundCommand result = new CompoundCommand();

        LayoutCalculator layoutCalculator = new LayoutCalculator();
        for (final IPathNotification notification : notifications) {
            final Object notifier = notification.getNotifier();
            if (processed.add(notifier)) {
                STAMPEdge edge = (STAMPEdge) notification.getNotifier();
                EObject element = DiagramElements.getModelElement(edge);
                if (!(element instanceof Link)) {
                    continue;
                }
                Link link = (Link) element;
                List<DiagramElement> syncedLabels = FluentIterable.from(link.getSignals())
                        .transform(signal -> resolveOrCreateLabel(signal, edge)).toList();
                STAMPLabel lastLabel = null;
                Location lastLocation = null;
                for (DiagramElement syncedLabel : syncedLabels) {
                    STAMPLabel label = (STAMPLabel) syncedLabel;
                    if (lastLabel == null || label.getLocation() != null) {
                        lastLabel = label;
                        continue;
                    }
                    IItemGeometryProvider provider = ItemProviders.adapt(domain, lastLabel,
                            IItemGeometryProvider.class);
                    if (provider == null) {
                        continue;
                    }
                    layoutCalculator.getFeatureAccessor().put(lastLabel,
                            DIPackage.Literals.DIAGRAM_ELEMENT__OWNING_ELEMENT, edge);
                    Dimension size = provider.getBoundsSize(lastLabel, layoutCalculator);
                    LocationConstraint locationConstraint = lastLabel.getLocation();
                    if (locationConstraint != null) {
                        lastLocation = (Location) locationConstraint;
                    } else {
                        if (lastLocation == null) {
                            lastLocation = NotationFactory.eINSTANCE.createLocation(0, 0);
                        }
                    }
                    Location location = NotationFactory.eINSTANCE.createLocation(
                            lastLocation.getX(), lastLocation.getY() + size.getHeight() + 5);
                    Command command = SetCommand.create(domain, label,
                            NotationPackage.Literals.SHAPE__LOCATION, location);
                    result.appendIfCanExecute(command);
                    lastLabel = label;
                    lastLocation = location;
                }
                result.appendIfCanExecute(
                        new SynchronizeListCommand(domain, edge.getOwnedElements(), syncedLabels));
            }
        }

        return result.isEmpty() ? null : result.unwrap();
    }

    private DiagramElement resolveOrCreateLabel(Signal signal, STAMPEdge edge) {
        return Iterables
                .tryFind(edge.getOwnedElements(),
                        label -> label.getModelElements().contains(signal))
                .or(createLabel(signal, edge));
    }

    private STAMPLabel createLabel(Signal signal, STAMPEdge edge) {
        STAMPLabel label = STAMPNotationFactory.eINSTANCE.createSTAMPLabel();
        DiagramElements.setModelElement(label, signal);
        label.setText(signal.getName());
        return label;
    }

}
