package net.astah.stpa.stamp.notation.editor;

import static net.astah.dd.di.edit.provider.IItemGraphicalStyleProvider.Feature.FONT_COLOR;
import static org.eclipse.emf.edit.domain.AdapterFactoryEditingDomain.getEditingDomainFor;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.util.Arrays;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.edit.provider.IItemPropertyDescriptor;
import org.eclipse.emf.edit.provider.IItemPropertySource;

import net.astah.dd.dc.Bounds;
import net.astah.dd.di.DiagramElement;
import net.astah.dd.di.edit.provider.Geometries;
import net.astah.dd.di.edit.provider.LayoutCalculator;
import net.astah.dd.di.util.DiagramElements;
import net.astah.emf.common.ecore.EObjects;
import net.astah.emf.edit.provider.ItemProviders;
import net.astah.golf.control.GolfAction;
import net.astah.golf.control.KeybindHandler;
import net.astah.golf.draw.GGeometry;
import net.astah.golf.draw.GNode;
import net.astah.golf.draw.GRoundRect;
import net.astah.golf.geom2D.Pnt2d;
import net.astah.golf.geom2D.Rectangle2d;
import net.astah.jomt.jutil.JomtWidgetFactory;
import net.astah.jomt.jview.BaseHighlighter;
import net.astah.jomt.jview.Highlighter;
import net.astah.jomt.jviewElement.CJomtTextNode;
import net.astah.notation.Shape;
import net.astah.notation.edit.layout.Layout;
import net.astah.notation.edit.layout.TextAreaLayout;
import net.astah.notation.edit.provider.IItemLayoutProvider;
import net.astah.notation.editor.GraphicalStyles;
import net.astah.notation.editor.IEdgeViewElement;
import net.astah.notation.editor.IMovableViewElement;
import net.astah.notation.editor.IResizableViewElement;
import net.astah.notation.editor.ITransformer;
import net.astah.notation.editor.IViewElement;
import net.astah.notation.editor.ViewElements;
import net.astah.notation.editor.adjuster.MovingAdjuster;
import net.astah.stpa.stamp.ControlAction;
import net.astah.stpa.stamp.Element;
import net.astah.stpa.stamp.Feedback;
import net.astah.stpa.stamp.Link;
import net.astah.stpa.stamp.ProcessModel;
import net.astah.stpa.stamp.ProcessValue;
import net.astah.stpa.stamp.ProcessVariable;
import net.astah.stpa.stamp.Signal;
import net.astah.stpa.stamp.notation.STAMPCompartment;
import net.astah.stpa.stamp.notation.STAMPComponentShape;
import net.astah.stpa.stamp.notation.STAMPDiagramElement;
import net.astah.stpa.stamp.notation.STAMPEdge;
import net.astah.stpa.stamp.notation.STAMPLabel;
import net.astah.stpa.stamp.notation.edit.layout.LayoutUtil;

/**
 * ラベル要素を表示するためのビューエレメントを実装するクラスです。
 */
public class STAMPLabelViewElement extends STAMPShapeViewElementImpl<STAMPLabel>
        implements IMovableViewElement, IResizableViewElement, KeybindHandler {

    private CJomtTextNode name;
    private MovingAdjuster.Factory movingAdjusterFactory;

    private Highlighter highlighter;

    private static final byte KNOB_COLOR = BLACK;
    private static final byte NAME_COLOR = BLACK;

    private static final String ACTION_PREFIX = "net.astah.stpa.stamp.ui.control.";
    private static final String ACTION_POSTFIX = "%%%d";

    private static final String ACTION_ADD_CONTROL_ACTION = ACTION_PREFIX + "AddControlAction"
            + ACTION_POSTFIX;
    private static final String ACTION_ADD_FEEDBACK = ACTION_PREFIX + "AddFeedback"
            + ACTION_POSTFIX;
    private static final String ACTION_ADD_PROCESS_VARIABLE = ACTION_PREFIX + "AddProcessVariable"
            + ACTION_POSTFIX;
    private static final String ACTION_ADD_PROCESS_VALUE = ACTION_PREFIX + "AddProcessValue"
            + ACTION_POSTFIX;

    public STAMPLabelViewElement(final STAMPLabel model) {
        super(model);
    }

    @Override
    public void refresh() {
        super.refresh();
        init();
    }

    private void init() {
        initName();
        initAdjusterFactory();
        initHighlighter();
    }

    private void initHighlighter() {
        highlighter = new BaseHighlighter(this);
    }

    @Override
    public <T> T getAdapter(Class<T> adapter) {
        if (adapter.isInstance(movingAdjusterFactory)) {
            return adapter.cast(movingAdjusterFactory);
        } else if (adapter.isInstance(highlighter)) {
            IEdgeViewElement edgeViewElement = getAncestor(IEdgeViewElement.class);
            STAMPComponentViewElement componentViewElement = getAncestor(
                    STAMPComponentViewElement.class);
            if (edgeViewElement == null && componentViewElement != null) {
                return componentViewElement.getAdapter(adapter);
            }
            return adapter.cast(highlighter);
        }
        return super.getAdapter(adapter);
    }

    private void initName() {
        if (name == null) {
            name = JomtWidgetFactory.instance().createJomtTextAreaNode();
            if (getParent() instanceof STAMPLinkViewElement
                    || ProcessModelElements.isProcessModelElement(getModel())) {
                name.setType(TEXT | RECT | BODY);
                name.setWrapLine(true);
            } else {
                if (getParent() instanceof STAMPComponentViewElement) {
                    name.setType(TEXT | RECT);
                } else {
                    name.setType(TEXT);
                }
                name.setWrapLine(isMultiline());
            }
            name.setModel(getModel());

            if (getParent() instanceof STAMPNoteViewElement || isProcessModelName()) {
                name.setSelectionType(GNode.NONE);
                name.setSelectionKnobNodeType(CJomtTextNode.NONE);
                name.setComponentSelectAll(false);
            } else {
                name.setSelectionType(GNode.KNOB);
                name.setSelectionKnobNodeType(CJomtTextNode.RECT);
                name.setComponentSelectAll(true);
                name.setAllowedNewLine(false);
            }

            name.addNodeSelectionListener(this);
            name.setAlpha(0.0f);
            final STAMPDiagramElement owner = getModel().getSTAMPOwningElement();
            if (!(owner instanceof STAMPEdge)) {
                final TextAreaLayout layout = getTextAreaLayout();
                if (layout != null && layout.isLineWrappable() && !layout.isWidthStretch()) {
                    addFocusListener(owner, layout);
                }
            }
            if (isProcessModelName()) {
                name.setEditable(false);
            }
            addNode(name);
        }
    }

    private void initAdjusterFactory() {
        final STAMPCompartmentViewElement viewElement = ViewElements.getAncestor(this,
                STAMPCompartmentViewElement.class);
        if (viewElement != null) {
            movingAdjusterFactory = viewElement.getAdapter(MovingAdjuster.Factory.class);
        }
    }

    private boolean isMultiline() {
        final Element element = DiagramElements.getModelElement(getModel(), Element.class);
        if (element == null) {
            return false;
        }
        final IItemPropertySource propertySource = ItemProviders.adapt(element,
                IItemPropertySource.class);
        if (propertySource != null) {
            for (final String featureName : Arrays.asList("name", "body")) {
                final IItemPropertyDescriptor propertyDescriptor = propertySource
                        .getPropertyDescriptor(element, featureName);
                if (propertyDescriptor == null) {
                    continue;
                }
                final Object feature = propertyDescriptor.getFeature(element);
                if (feature instanceof EStructuralFeature && ((EStructuralFeature) feature)
                        .getEType().getInstanceClass() == String.class) {
                    return propertyDescriptor.isMultiLine(element);
                }
            }
        }
        return false;
    }

    private void addFocusListener(final STAMPDiagramElement owner, final TextAreaLayout layout) {
        final Component component = name.getTextComponent().getComponent();
        component.addFocusListener(new FocusAdapter() {
            @Override
            public void focusGained(final FocusEvent e) {
                if (getEditingDomainFor(getModel()) != null) {
                    setWidthToComponent(component, owner, layout);
                }
            }
        });
    }

    private void setWidthToComponent(final Component component, final STAMPDiagramElement owner,
            final TextAreaLayout layout) {
        if (layout.needsToWrapLine(getModel())) {
            component.setSize(
                    (int) (new LayoutCalculator().getBounds(owner).getWidth()
                            - layout.getLeftMargin() - layout.getRightMargin()),
                    component.getHeight());
        }
    }

    @Override
    protected void update() {
        super.update();
        updateName();
    }

    private void updateName() {
        String mark = "";
        final EObject element = DiagramElements.getModelElement(getModel());
        if (element instanceof ProcessVariable) {
            mark = "* ";
        } else if (element instanceof ProcessValue) {
            mark = "- ";
        }
        final String nameText = getModel().getText() == null ? "" : getModel().getText();
        name.setString(mark + nameText);
        name.setSelectedString(nameText);
    }

    @Override
    public void updateStyle() {
        super.updateStyle();
        updateFont();
    }

    @Override
    protected void updateColor() {
        GraphicalStyles.applyFontColor(name, getModel(), NAME_COLOR);
        name.setKnobColorNo(KNOB_COLOR);
    }

    @Override
    public boolean isHighlighted() {
        if (getModel().getOwningElement() instanceof STAMPComponentShape) {
            return false;
        }
        return super.isHighlighted();
    }

    @Override
    protected GNode createHighlightShape() {
        GRoundRect rect = createHighlightRect();
        rect.setAlpha(0.2f);
        rect.setFilled(true);
        rect.setBackgroundColor(GraphicalStyles.getHighlightColor(getModel(), FONT_COLOR));
        return rect;
    }

    private GRoundRect createHighlightRect() {
        Rectangle2d bounds = new Rectangle2d();
        getBounds(bounds);
        double x = bounds.x - HIGH_LIGHT_MARGIN;
        double y = bounds.y - HIGH_LIGHT_MARGIN;
        double width = bounds.width + 2 * HIGH_LIGHT_MARGIN;
        double height = bounds.height + 2 * HIGH_LIGHT_MARGIN;
        return new GRoundRect(x, y, width, height, HIGHT_LIGHT_ARC, HIGHT_LIGHT_ARC);
    }

    /**
     * フォントの更新処理を行います。
     */
    protected void updateFont() {
        GraphicalStyles.applyFont(name, getModel());
    }

    @Override
    protected void layout(final Bounds bounds) {
        super.layout(bounds);
        final TextAreaLayout layout = getTextAreaLayout();
        if (layout != null) {
            name.setWrapLine(layout.needsToWrapLine(getModel()));
        }
        layoutNameRect(bounds);
    }

    @Override
    protected void layoutHighlightShape() {
        super.layoutHighlightShape();

        if (highlight != null) {
            GRoundRect rect = createHighlightRect();
            ((GRoundRect) highlight).setRoundRect(rect.getX(), rect.getY(), rect.getWidth(),
                    rect.getHeight());
        }
    }

    private void layoutNameRect(final Bounds bounds) {
        if (bounds.getHeight() != 0) {
            name.setRect(bounds.getX(), bounds.getY(), bounds.getWidth(), bounds.getHeight());
        } else {
            // TextAreaLayout::skipによりレイアウト処理結果の高さが0になったテキストは表示しない
            name.setVisible(false);
        }
    }

    private boolean isProcessModelName() {
        final STAMPLabel label = getModel();
        final DiagramElement owner = label.getOwningElement();
        if (!(owner instanceof STAMPCompartment)) {
            return false;
        }
        final EObject element = DiagramElements.getModelElement(owner);
        if (!(element instanceof ProcessModel)) {
            return false;
        }
        final EList<DiagramElement> elements = owner.getOwnedElements();
        return !elements.isEmpty() && elements.get(0) == label;
    }

    /**
     * このViewElementに紐づく図要素のレイアウトに使われるTextAreaLayoutを取得します。
     * 
     * @return このViewElementに紐づく図要素のレイアウトに使われるTextAreaLayout
     */
    public TextAreaLayout getTextAreaLayout() {
        final STAMPDiagramElement owner = getModel().getSTAMPOwningElement();
        final IItemLayoutProvider provider = ItemProviders.adapt(owner, IItemLayoutProvider.class);
        final Layout layout = provider.getLayout(owner);
        if (layout == null) {
            return null;
        }
        final Layout ownedLayout = layout.getOwnedLayoutById(LayoutUtil.TEXT_AREA);
        return ownedLayout instanceof TextAreaLayout ? (TextAreaLayout) ownedLayout : null;
    }

    @Override
    protected GGeometry getBounds(Rectangle2d bounds) {
        return super.getBounds(name.body, bounds);
    }

    @Override
    public IViewElement getMovableViewElement() {
        final GNode parent = getParent();
        if (parent instanceof IViewElement) {
            final IViewElement movableViewElement = ((IViewElement) parent).getMovableViewElement();
            if (movableViewElement != null) {
                return movableViewElement;
            }
        }
        return super.getMovableViewElement();
    }

    @Override
    public GNode createMovingShape() {
        final GNode node = super.createMovingShape();
        if (node == null) {
            return null;
        }
        final GNode parent = getParent();
        if (parent instanceof IViewElement) {
            final GNode tempNode = ((IViewElement) parent).createMovingShape();
            if (tempNode != null) {
                node.addChild(tempNode);
            }
        }
        return node;
    }

    @Override
    public void setDepth(int depth) {
        super.setDepth(depth);
        name.setDepth(depth);
    }

    @Override
    public void setSelected() {
        if (!(getParent() instanceof STAMPComponentViewElement)) {
            super.setSelected();
        }
    }

    @Override
    public void selectionChange(boolean selected) {
        if (getParent() instanceof STAMPNoteViewElement) {
            getParent().setSelected(selected);
        }
    }

    @Override
    public ITransformer startMove(Pnt2d startLocation) {
        final GNode parentNode = getParent();
        if (parentNode instanceof STAMPComponentViewElement) {
            return ((STAMPComponentViewElement) parentNode).startMove(startLocation);
        }
        if (parentNode instanceof STAMPNoteViewElement) {
            return null;
        }
        return new STAMPLabelTransformer(this, startLocation);
    }

    @Override
    public ITransformer startResize(Pnt2d startLocation, int knob) {
        final GNode parentNode = getParent();
        if (parentNode instanceof STAMPComponentViewElement
                || parentNode instanceof STAMPNoteViewElement) {
            return null;
        }
        return new STAMPLabelTransformer(this, startLocation, knob);
    }

    @Override
    public void handleKeybind(Object action) {
        String actionName = getActionName(action);
        if (actionName == null) {
            return;
        }
        GolfAction golfAction = new GolfAction(actionName);
        ActionEvent actionEvent = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, actionName);
        golfAction.actionPerformed(actionEvent);
    }

    private String getActionName(Object action) {
        EObject model = DiagramElements.getModelElement(getModel());
        if (action == ActionKind.PRESSED_ENTER || action == ActionKind.PRESSED_SHIFT_ENTER) {
            if (model instanceof Signal) {
                Link link = EObjects.getAncestor(Link.class, model);
                int index = link.getSignals().indexOf(model);
                if (action == ActionKind.PRESSED_ENTER) {
                    index++;
                }
                if (model instanceof ControlAction) {
                    return String.format(ACTION_ADD_CONTROL_ACTION, index);
                }
                if (model instanceof Feedback) {
                    return String.format(ACTION_ADD_FEEDBACK, index);
                }
            } else if (model instanceof ProcessVariable) {
                final ProcessModel processModel = EObjects.getAncestor(ProcessModel.class, model);
                int index = processModel.getVariables().indexOf(model);
                if (action == ActionKind.PRESSED_ENTER) {
                    index++;
                }
                return String.format(ACTION_ADD_PROCESS_VARIABLE, index);
            } else if (model instanceof ProcessValue) {
                final ProcessVariable processVariable = EObjects.getAncestor(ProcessVariable.class,
                        model);
                int index = processVariable.getValues().indexOf(model);
                if (action == ActionKind.PRESSED_ENTER) {
                    index++;
                }
                return String.format(ACTION_ADD_PROCESS_VALUE, index);
            }
        }
        return null;
    }

    /**
     * 移動(順序変更)先のラベルを取得します。
     * 
     * @param end
     *            移動先の点
     * @return 移動先のラベル<br>
     *         移動先が不正な場合はnull
     */
    public Shape getDestinationLabelToReorder(final Pnt2d end) {
        final DiagramElement compartment = getModel().getOwningElement();
        final Shape owner = (Shape) compartment.getOwningElement();
        final Rectangle2d rect = Geometries.toRectangle2d(owner.getBounds());
        if (!rect.contains(end)) {
            return null;
        }
        final EList<DiagramElement> diagramElements = compartment.getOwnedElements();
        Shape label = null;
        // 区画の名前は対象外
        for (int i = 1; i < diagramElements.size(); i++) {
            label = (Shape) diagramElements.get(i);
            final Bounds bounds = label.getBounds();
            if (end.y <= bounds.getY() + bounds.getHeight()) {
                return label;
            }
        }
        return label;
    }

    @Override
    public int compareTo(IViewElement viewElement) {
        final STAMPLabel label = getModel();
        if (!ProcessModelElements.isProcessModelElement(label)) {
            return super.compareTo(viewElement);
        }
        if (!(viewElement instanceof STAMPLabelViewElement)) {
            return super.compareTo(viewElement);
        }
        final STAMPLabel comparedLabel = ((STAMPLabelViewElement) viewElement).getModel();
        if (!ProcessModelElements.isProcessModelElement(comparedLabel)) {
            return super.compareTo(viewElement);
        }
        // プロセスモデル要素のViewElementはSTAMPLabelTransformerマージ後のターゲットリストtargets内でy座標の昇順に並ぶようにする
        return Double.compare(label.getBounds().getY(), comparedLabel.getBounds().getY());
    }
}
