package net.astah.stpa.stamp.notation.editor;

import static net.astah.dd.di.edit.provider.IItemGraphicalStyleProvider.Feature.FILL_COLOR;
import static net.astah.dd.di.edit.provider.IItemGraphicalStyleProvider.Feature.STROKE_COLOR;
import static net.astah.dd.di.edit.provider.IItemGraphicalStyleProvider.Feature.STROKE_WIDTH;

import java.awt.Color;

import javax.swing.ImageIcon;

import org.eclipse.emf.common.util.URI;

import net.astah.dd.dc.Bounds;
import net.astah.dd.di.edit.provider.IItemGraphicalStyleProvider;
import net.astah.emf.edit.provider.ItemProviders;
import net.astah.emf.edit.provider.swing.ColorURIResolver;
import net.astah.golf.draw.GGeometry;
import net.astah.golf.draw.GNode;
import net.astah.golf.draw.GRect;
import net.astah.golf.geom2D.Pnt2d;
import net.astah.golf.geom2D.Rectangle2d;
import net.astah.golf.widget.CImageIcon;
import net.astah.golf.widget.GolfWidgetFactory;
import net.astah.golf.widget.swing.SwingColor;
import net.astah.golf.widget.swing.SwingImageIcon;
import net.astah.jomt.jview.HyperLinkIcon;
import net.astah.jomt.jview.HyperLinkIconImpl;
import net.astah.jomt.jview.IconResource;
import net.astah.jomt.jview.RectangleDrawSuggest;
import net.astah.jomt.jviewElement.CommentBody;
import net.astah.notation.editor.IDiagramViewElement;
import net.astah.notation.editor.IMovableViewElement;
import net.astah.notation.editor.IResizableViewElement;
import net.astah.notation.editor.ITransformer;
import net.astah.notation.editor.ShapeTransformer;
import net.astah.notation.editor.ViewElements;
import net.astah.notation.editor.adjuster.MovingAdjuster;
import net.astah.notation.editor.adjuster.ResizingAdjuster;
import net.astah.stpa.stamp.notation.STAMPShape;

/**
 * ノート要素を表示するためのビューエレメントを実装するクラスです。
 */
public class STAMPNoteViewElement extends STAMPShapeViewElementImpl<STAMPShape>
        implements IMovableViewElement, IResizableViewElement {
    // color.
    static final byte BODY_COLOR = WHITE;
    static final byte FRAME_COLOR = BLACK;
    static final byte KNOB_COLOR = BLACK;
    static final byte SHADOW_COLOR = LIGHTGRAY;
    static final byte CHAR_COLOR = BLACK;

    /** 本体ノード */
    protected CommentBody commentBody;

    /** フレームノード */
    protected GRect frame;

    private RectangleDrawSuggest drawSuggest;
    private HyperLinkIcon hyperLinkIcon;
    private MovingAdjuster.Factory movingAdjusterFactory;
    private ResizingAdjuster.Factory resizingAdjusterFactory;

    public STAMPNoteViewElement(final STAMPShape model) {
        super(model);
    }

    @Override
    protected void refresh() {
        super.refresh();
        init();
    }

    private void init() {
        initFrame();
        initCommentBody();
        initDrawSuggest();
        initHyperLinkIcon();
        initAdjusterFactory();
    }

    @Override
    public <T> T getAdapter(Class<T> adapter) {
        if (adapter.isInstance(drawSuggest)) {
            return adapter.cast(drawSuggest);
        } else if (adapter.isInstance(hyperLinkIcon)) {
            return adapter.cast(hyperLinkIcon);
        } else if (adapter.isInstance(movingAdjusterFactory)) {
            return adapter.cast(movingAdjusterFactory);
        } else if (adapter.isInstance(resizingAdjusterFactory)) {
            return adapter.cast(resizingAdjusterFactory);
        }
        return super.getAdapter(adapter);
    }

    private void initFrame() {
        if (frame == null) {
            frame = new GRect();
            int type = 0;
            type = type | RECT;
            type = type | BODY;
            frame.setType(type);
            frame.setModel(getModel());
            frame.setFilled(false);
            frame.setAlpha(0.0f);
            frame.setSelectionType(GNode.KNOB);
            frame.addNodeSelectionListener(frame);
            frame.addNodeSelectionListener(this);
            addNodeSelectionListener(frame);
            addNode(frame);
        }
    }

    private void initCommentBody() {
        if (commentBody == null) {
            commentBody = new CommentBody();
            commentBody.setModel(getModel());
            commentBody.setBackgroundNo(BODY_COLOR);
            commentBody.setFilled(true);
            commentBody.setSelectionType(GNode.SHADOW);
            frame.addNodeSelectionListener(commentBody);
            addNode(commentBody);
        }
    }

    private void initDrawSuggest() {
        drawSuggest = new RectangleDrawSuggest(this);

        CImageIcon image = new SwingImageIcon(
                (ImageIcon) IconResource.getNormalSizeNoteAnchorIcon());
        drawSuggest.addModeAction("net.astah.stpa.stamp.ui.control.mode.CreateNoteAnchor", image);
    }

    private void initHyperLinkIcon() {
        if (hyperLinkIcon != null) {
            return;
        }
        hyperLinkIcon = new HyperLinkIconImpl(this, getModel());
        hyperLinkIcon.addIconNode();
    }

    private void initAdjusterFactory() {
        movingAdjusterFactory = ViewElements.getAncestor(this, IDiagramViewElement.class)
                .getAdapter(MovingAdjuster.Factory.class);
        resizingAdjusterFactory = ViewElements.getAncestor(this, IDiagramViewElement.class)
                .getAdapter(ResizingAdjuster.Factory.class);
    }

    @Override
    protected void update() {
        super.update();
        updateHyperLinkIcon();
    }

    private void updateHyperLinkIcon() {
        hyperLinkIcon.update();
    }

    @Override
    protected void updateStroke() {
        final Object model = getModel();
        final IItemGraphicalStyleProvider provider = ItemProviders.adapt(model,
                IItemGraphicalStyleProvider.class);
        final Object width = provider.getGraphicalStyle(model, STROKE_WIDTH);
        if (width instanceof Number) {
            commentBody.setRimWidth(((Number) width).intValue());
        } else {
            commentBody.setRimWidth(1);
        }
    }

    @Override
    public void updateColor() {
        final Object model = getModel();
        final IItemGraphicalStyleProvider provider = ItemProviders.adapt(model,
                IItemGraphicalStyleProvider.class);
        final URI fillColorURI = (URI) provider.getGraphicalStyle(model, FILL_COLOR);
        if (fillColorURI != null) {
            final Color fillColor = new ColorURIResolver(null, null).resolve(fillColorURI);
            commentBody.setBackgroundColor(new SwingColor(fillColor));
        } else {
            commentBody.setBackgroundColor(GolfWidgetFactory.instance().createCColor(Color.PINK));
        }
        final URI strokeColorURI = (URI) provider.getGraphicalStyle(model, STROKE_COLOR);
        if (strokeColorURI != null) {
            final Color strokeColor = new ColorURIResolver(null, null).resolve(strokeColorURI);
            commentBody.setForegraundColor(new SwingColor(strokeColor));
        } else {
            commentBody.setColorNo(FRAME_COLOR);
        }

        frame.setKnobColorNo(KNOB_COLOR);
        commentBody.setShadowColorNo(SHADOW_COLOR);
    }

    @Override
    protected void layout(final Bounds bounds) {
        super.layout(bounds);
        layoutFrame(bounds);
        layoutHyperLinkIcon();
    }

    /**
     * ハイパーリンクアイコンのレイアウト計算を行います。
     */
    protected void layoutHyperLinkIcon() {
        hyperLinkIcon.relocate(frame);
    }

    private void layoutFrame(final Bounds bounds) {
        frame.setRect(bounds.getX(), bounds.getY(), bounds.getWidth(), bounds.getHeight());
        commentBody.update(bounds.getX(), bounds.getY(), bounds.getWidth(), bounds.getHeight());
    }

    @Override
    protected GGeometry getBounds(final Rectangle2d bounds) {
        return super.getBounds(frame, bounds);
    }

    @Override
    public void setDepth(int depth) {
        super.setDepth(depth);
        commentBody.setDepth(depth);
        frame.setDepth(depth - 1);
    }

    @Override
    public void setSelected(boolean selected) {
        frame.setSelected(selected);
        super.setSelected(selected);
    }

    @Override
    public ITransformer startMove(Pnt2d startLocation) {
        return new ShapeTransformer(this, startLocation);
    }

    @Override
    public ITransformer startResize(Pnt2d startLocation, int knob) {
        return new ShapeTransformer(this, startLocation, knob);
    }
}
