package net.astah.stpa.stamp.ui.control;

import static java.util.Collections.singleton;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.command.UnexecutableCommand;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.edit.command.CommandParameter;
import org.eclipse.emf.edit.command.CreateChildCommand;
import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.eclipse.emf.transaction.util.TransactionUtil;

import net.astah.dd.dc.Bounds;
import net.astah.dd.di.DIPackage;
import net.astah.dd.di.DiagramElement;
import net.astah.dd.di.Shape;
import net.astah.dd.di.util.DiagramElements;
import net.astah.emf.common.ecore.EObjects;
import net.astah.emf.edit.job.CommandJob;
import net.astah.emf.edit.job.Jobs;
import net.astah.jomt.jsystem.JomtGlobal;
import net.astah.jomt.jview.DiagramEditor;
import net.astah.jomt.jview.DiagramEditors;
import net.astah.notation.Location;
import net.astah.notation.NotationFactory;
import net.astah.notation.Size;
import net.astah.notation.editor.ViewElements;
import net.astah.stpa.stamp.Link;
import net.astah.stpa.stamp.Signal;
import net.astah.stpa.stamp.notation.STAMPEdge;
import net.astah.stpa.stamp.notation.STAMPLabel;
import net.astah.stpa.stamp.notation.STAMPNotationFactory;

/**
 * シグナル（ コントロールアクション 、フィードバック ）を作成するコマンドの基底クラスです。
 */
public abstract class AddSignalCommand extends net.astah.golf.control.Command {

    private Shape shape;

    private STAMPEdge parent;

    /** シフトボタンがおされているかどうか */
    private boolean isShiftPressed = false;

    private int index = -1;

    public AddSignalCommand() {
    }

    @Override
    public void setArgumentString(String arg) {
        try {
            this.index = Integer.valueOf(arg);
        } catch (NumberFormatException e) {
            // 例外が起こる時デフォルトの値 index=-1 を利用する
        }
    }

    public void setShape(Shape shape) {
        this.shape = shape;
    }

    public void setParent(STAMPEdge parent) {
        this.parent = parent;
    }

    /**
     * Shift キーが押されているかの状態を保存します。
     * 
     * @param isShiftPressed
     *            true : Shift キーが押されている。 false : Shift キーが押されていない。
     */
    public void setShiftPressedState(boolean isShiftPressed) {
        this.isShiftPressed = isShiftPressed;
    }

    @Override
    public void execute() {
        if (parent == null) {
            parent = getParent();
        }
        if (!isValidTarget(parent)) {
            return;
        }

        TransactionalEditingDomain domain = TransactionUtil.getEditingDomain(parent);

        if (shape == null) {
            shape = STAMPNotationFactory.eINSTANCE.createSTAMPLabel();
            DiagramElements.setModelElement(shape, createSignal());
        }

        STAMPLabel lastLabel = getLastAddLinkLabel();
        if (lastLabel != null) {
            // 他にリンクのラベルがある場合は、最後に追加したラベルの下に移動させる
            Location location = (Location) lastLabel.getLocation();
            if (location == null) {
                location = NotationFactory.eINSTANCE.createLocation(0, 0);
            }

            Bounds bounds = lastLabel.getBounds();
            Size size = NotationFactory.eINSTANCE.createSize(bounds.getWidth(), bounds.getHeight());
            if (location != null) {
                ((STAMPLabel) shape).setLocation(
                        NotationFactory.eINSTANCE.createLocation(location.getX(),
                                location.getY() + size.getHeight() + 5));
            }
        }

        // このコマンドが実行するEMFコマンドを構築する
        Command command = createTransactionalCommand(domain);

        // 構築したEMFコマンドを非同期ジョブとして実行する
        CommandJob job = new CommandJob(domain, command);
        Jobs.toCompletableFuture(job, true).thenRun(this::postExecute);
        job.schedule();
    }

    private boolean isValidTarget(STAMPEdge edge) {
        return edge != null && DiagramElements.isModelElementPresent(edge, getLinkType());
    }

    protected abstract Signal createSignal();

    protected abstract Class<? extends Link> getLinkType();

    private STAMPLabel getLastAddLinkLabel() {
        EList<DiagramElement> elements = parent.getOwnedElements();
        STAMPLabel lastLabel = null;
        for (DiagramElement element : elements) {
            if (element instanceof STAMPLabel && element != shape) {
                lastLabel = (STAMPLabel) element;
            }
        }
        return lastLabel;
    }

    private STAMPEdge getParent() {
        DiagramElement[] targets = null;
        DiagramEditor editor = JomtGlobal.viewManager.getCurrentDiagramEditor();
        if (editor != null) {
            targets = editor.getSelectedPresentations();
        }
        if (targets != null) {
            for (DiagramElement target : targets) {
                STAMPEdge edge = EObjects.getAncestor(STAMPEdge.class, target);
                if (edge != null) {
                    return edge;
                }
            }
        }
        return null;
    }

    /**
     * トランザクションとして実行するモデル編集コマンドを作成します。
     * 
     * @param domain
     *            コマンドを実行するドメイン
     * @return 作成したコマンド
     */
    public Command createTransactionalCommand(TransactionalEditingDomain domain) {
        if (shape == null || parent == null) {
            return UnexecutableCommand.INSTANCE;
        }

        return CreateChildCommand.create(domain, parent,
                new CommandParameter(parent,
                        DIPackage.Literals.DIAGRAM_ELEMENT__OWNED_ELEMENT, shape, index),
                singleton(parent));
    }

    /**
     * トランザクション完了後の後処理を行います。
     */
    protected void postExecute() {
        if (JomtGlobal.viewManager != null) {
            DiagramEditor currentDiagramEditor = JomtGlobal.viewManager.getCurrentDiagramEditor();
            if (currentDiagramEditor != null && currentDiagramEditor.getCanvas() != null) {
                currentDiagramEditor.getCanvas().unselect();
                DiagramEditors.selectOn(currentDiagramEditor, shape);
                ViewElements.startEditing(currentDiagramEditor.getCanvas(), shape);
            }
        }

        if (JomtGlobal.modeManager != null) {
            JomtGlobal.modeManager.changeButtonsState(isShiftPressed);
        }
    }

}
