package net.astah.stpa.stamp.ui.menu.handler;

import java.awt.event.ActionEvent;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

import org.eclipse.emf.common.command.CompoundCommand;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.edit.command.AddCommand;
import org.eclipse.emf.edit.command.CommandParameter;
import org.eclipse.emf.edit.command.CreateChildCommand;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.eclipse.emf.transaction.util.TransactionUtil;

import net.astah.dd.di.util.DiagramElements;
import net.astah.emf.edit.command.DestroyCommand;
import net.astah.emf.edit.job.CommandJob;
import net.astah.golf.widget.ISelection;
import net.astah.golf.widget.swing.Selections;
import net.astah.stpa.stamp.Component;
import net.astah.stpa.stamp.ProcessModel;
import net.astah.stpa.stamp.STAMPFactory;
import net.astah.stpa.stamp.STAMPPackage;
import net.astah.stpa.stamp.notation.STAMPCompartment;
import net.astah.stpa.stamp.notation.STAMPComponentShape;
import net.astah.stpa.stamp.notation.STAMPNotationPackage;
import net.astah.stpa.stamp.notation.edit.command.SynchronizeProcessModelCommand;
import net.astah.ui.menu.annotation.CanExecute;
import net.astah.ui.menu.annotation.Execute;
import net.astah.ui.menu.annotation.Handler;
import net.astah.ui.menu.annotation.MenuCommand;
import net.astah.ui.menu.annotation.MenuItem;
import net.astah.ui.menu.annotation.MenuItem.ItemKind;

/**
 * 複数選択時にプロセスモデル区画の表示非表示を切り替えます。
 */
@Handler(commands = "stpa.ShowPluralProcessModelCompartmentCommand")
@MenuItem(commands = @MenuCommand(id = "stpa.ShowPluralProcessModelCompartmentCommand",
        category = "edit.show_diagram_element"),
        path = "ShowPluralProcessModelCompartment",
        label = "%StampControlStructureDiagram_showProcessModelCompartment_label",
        tooltip = "%StampControlStructureDiagram_showProcessModelCompartment_tooltip",
        type = ItemKind.MENU)
@MenuItem(commands = @MenuCommand(id = "stpa.ShowPluralProcessModelCompartmentCommand",
        category = "edit.show_diagram_element",
        parameters = ShowPluralProcessModelCompartmentHandler.ON),
        path = "ShowPluralProcessModelCompartment.on",
        label = "%StampControlStructureDiagram_showPluralProcessModelCompartment_on_label",
        tooltip = "%StampControlStructureDiagram_showPluralProcessModelCompartment_on_tooltip",
        type = ItemKind.PUSH)
@MenuItem(commands = @MenuCommand(id = "stpa.ShowPluralProcessModelCompartmentCommand",
        category = "edit.show_diagram_element",
        parameters = ShowPluralProcessModelCompartmentHandler.OFF),
        path = "ShowPluralProcessModelCompartment.off",
        label = "%StampControlStructureDiagram_showPluralProcessModelCompartment_off_label",
        tooltip = "%StampControlStructureDiagram_showPluralProcessModelCompartment_off_tooltip",
        type = ItemKind.PUSH)
public class ShowPluralProcessModelCompartmentHandler {
    static final String ON = "on";
    static final String OFF = "off";

    /**
     * 複数選択時にプロセスモデル区画の表示非表示を切り替えるコマンドを実行します。
     * 
     * @param selection
     *            選択要素
     * @param params
     *            コマンドのパラメータ
     * @param e
     *            アクションイベント
     */
    @Execute
    public void execute(final ISelection selection, final List<String> params,
            final ActionEvent e) {
        final List<STAMPComponentShape> shapes = Selections.stream(selection)
                .filter(o -> isTarget(o)).map(it -> (STAMPComponentShape) it)
                .collect(Collectors.toList());
        final EObject element = DiagramElements.getModelElement(shapes.get(0));
        final TransactionalEditingDomain domain = TransactionUtil.getEditingDomain(element);
        final CompoundCommand command = new CompoundCommand();
        for (final STAMPComponentShape shape : shapes) {
            final Component component = DiagramElements.getModelElement(shape, Component.class);
            final ProcessModel processModel = component.getProcessModel();
            final STAMPCompartment compartment = Compartments
                    .getProcessModelCompartment(shape.getCompartments(), processModel);
            if (ON.equals(params.get(0))) {
                if (compartment == null) {
                    createCreateCompartmentCommand(domain, component, processModel, shape, command);
                }
            } else {
                if (compartment != null) {
                    createDeleteCompartmentCommand(domain, compartment, command);
                }
            }
        }
        final CommandJob job = new CommandJob(domain, command);
        job.schedule();
    }

    private void createCreateCompartmentCommand(final EditingDomain domain,
            final Component component, final ProcessModel existingProcessModel,
            final STAMPComponentShape shape, final CompoundCommand command) {
        if (existingProcessModel == null) {
            final ProcessModel processModel = STAMPFactory.eINSTANCE.createProcessModel();
            command.append(CreateChildCommand.create(
                    domain, component, new CommandParameter(component,
                            STAMPPackage.Literals.COMPONENT__PROCESS_MODEL, processModel),
                    Collections.singleton(component)));
            final STAMPCompartment compartment = Compartments.createCompartment(processModel);
            command.append(AddCommand.create(domain, shape,
                    STAMPNotationPackage.Literals.STAMP_COMPARTMENTABLE_SHAPE__COMPARTMENT,
                    compartment));
        } else {
            final STAMPCompartment compartment = Compartments
                    .createCompartment(existingProcessModel);
            command.append(AddCommand.create(domain, shape,
                    STAMPNotationPackage.Literals.STAMP_COMPARTMENTABLE_SHAPE__COMPARTMENT,
                    compartment));
            command.appendIfCanExecute(new SynchronizeProcessModelCommand(domain, compartment));
        }
    }

    private void createDeleteCompartmentCommand(final EditingDomain domain,
            final STAMPCompartment compartment, final CompoundCommand command) {
        command.append(DestroyCommand.create(domain, compartment));
    }

    /**
     * 複数選択時にプロセスモデル区画の表示非表示を切り替えるコマンドが実行可能かどうか確認します。
     * 
     * @param selection
     *            選択要素
     * @return true コマンドが実行可能 / false コマンドが実行不可
     */
    @CanExecute
    public boolean canExecute(final ISelection selection) {
        return 2 <= Selections.size(selection)
                && Selections.stream(selection).anyMatch(o -> isTarget(o));
    }

    private boolean isTarget(final Object object) {
        return object instanceof STAMPComponentShape && !((STAMPComponentShape) object).isOmitted();
    }
}
